/*
    Copyright (C) 2004    John Orlando

   AVRcam: a small real-time image processing engine.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   For more information on the AVRcam, please contact:

   john@jrobot.net

   or go to www.jrobot.net for more details regarding the system.
*/
/**********************************************************
	Module Name: CamConfig.c
	Module Date: 04/10/2004
    Module Auth: John Orlando

	Description: This module is responsible for the
	high-level configuration activities of the OV6620
	camera module.  This module interfaces with the
	I2CInterface module to perform this configuration.

    Revision History:
    Date        Rel Ver.    Notes
    4/10/2004      0.1     Module created
    6/30/2004      1.0     Initial release for Circuit Cellar
                           contest.
    11/15/2004     1.2     Added code to un-tri-state the
                           OV6620's pixel data busses at
                           startup after four seconds.
                           This was added in to
                           allow the user to re-program the
                           mega8 at startup if needed.
***********************************************************/

/*  Includes */

/**********************************************************/
/*  Definitions */
/* The length of an I2C command is made up of a register address
plus the actual value of the register */
#define SIZE_OF_I2C_CMD 2
#define MAX_NUM_CONFIG_CMDS 8
#define CAM_CONFIG_TX_FIFO_SIZE MAX_NUM_CONFIG_CMDS
#define CAM_CONFIG_TX_FIFO_MASK CAM_CONFIG_TX_FIFO_SIZE-1

/* This is the I2C slave address of the OV6620 module.
NOTE: The actual address of the camera is 0xC0.  However,
the TWI interface on the mega8 needs to have the address
shifted one bit to the right, which yields 0x60.  The
LSB is used to set whether the operation is read or 
write. */
#define CAM_ADDRESS 0x60	

/*  Local Variables */

/*  Local Structures and Typedefs */

/*  Local Function Prototypes */

/*  Extern Variables */
I2CCommand 		CamConfigTxFifo[CAM_CONFIG_TX_FIFO_SIZE];
unsigned char CamConfigTxFifoHead=0;
unsigned char CamConfigTxFifoTail=0;

/***********************************************************
	Function Name: CamConfig_writeTxFifo
	Function Description: This function is responsible for
	adding a new command to the tx fifo.  It adjusts all
	needed pointers.
	Inputs:  cmd - the i2cCmd_t to add to the fifo
	Outputs: bool_t - indicating if writing to the fifo
	         causes it to wrap
***********************************************************/
unsigned char CamConfigWriteTxFifo(I2CCommand Cmd)
{
	unsigned char TempHead;
	unsigned char RetVal = TRUE;

	CamConfigTxFifo[CamConfigTxFifoHead] = Cmd;

	/* see if we need to wrap */
	TempHead = (CamConfigTxFifoHead+1) & (CAM_CONFIG_TX_FIFO_MASK);
	CamConfigTxFifoHead = TempHead;

	/* check to see if we have filled up the queue */
	if (CamConfigTxFifoHead == CamConfigTxFifoTail)
	{
		/* we wrapped the fifo...return false */
		RetVal = FALSE;
	}
	return(RetVal);
}

/***********************************************************
	Function Name: CamConfig_setCamReg
	Function Description: This function is responsible for
	creating an I2C cmd structure and placing it into the
	cmd fifo.
	Inputs:  reg - the register to modify
	         val - the new value of the register
	Outputs: none
***********************************************************/
void CamConfigSetCamReg(unsigned char Reg, unsigned char Value)
{
	I2CCommand Cmd;

	Cmd.ConfigReg = Reg;
	Cmd.Data = Value;
#ifndef SIMULATION
	CamConfigWriteTxFifo(Cmd);
#endif
}

/***********************************************************
	Function Name: CamConfig_readTxFifo
	Function Description: This function is responsible for
	reading a cmd out of the tx fifo.
	Inputs:  none
	Outputs: i2cCmd_t - the cmd read from the fifo
***********************************************************/
static I2CCommand CamConfigReadTxFifo(void)
{
	I2CCommand Cmd;
	unsigned char TempTail;

	/* just return the current tail from the rx fifo */
	Cmd = CamConfigTxFifo[CamConfigTxFifoTail];
	TempTail = (CamConfigTxFifoTail+1) & (CAM_CONFIG_TX_FIFO_MASK);
	CamConfigTxFifoTail = TempTail;

	return(Cmd);
}

/***********************************************************
	Function Name: CamConfig_sendFifoCmds
	Function Description: This function is responsible for
	sending the entire contents of the config fifo.  This
	function won't return until the configuration process
	is complete (or an error is encountered).
	Inputs:  none
	Outputs: none
	Note: Since this function is written to use the TWI
	interrupt in the I2CInterface module, there will be
	some busy-waiting here...no big deal, since we end up
	having to trash the frame that we are executing this
	slave write in anyway (since we can't meet the strict
	timing requirements and write i2c at the same time).
***********************************************************/
void CamConfigSendFIFOCmds(void)
{
	I2CCommand Cmd;

	while (CamConfigTxFifoHead != CamConfigTxFifoTail)
	{
		Cmd = CamConfigReadTxFifo();
		I2CWriteData(CAM_ADDRESS,&Cmd.ConfigReg,SIZE_OF_I2C_CMD);
		Utility_delay(100);
		/* wait for the I2C transaction to complete */
		while(IsI2CBusy() == TRUE);
	}
}

/***********************************************************
	Function Name: CamConfig_init
	Function Description: This function is responsible for
	performing the initial configuration of the camera.
	Inputs:  none
	Outputs: none
***********************************************************/
void CamConfigInit(void)
{
	CamConfigSetCamReg(0x14,0x20);  /* reduce frame size */
	CamConfigSetCamReg(0x39,0x40);  /* gate PCLK with HREF */
	CamConfigSetCamReg(0x12,0x28);  /* set RGB mode, with no AWB */
	CamConfigSetCamReg(0x28,0x05);  /* set color sequencer */
        CamConfigSetCamReg(0x13,0x01);  /* un-tri-state the Y/UV lines */

	/* send the first four cmds in the I2C fifo */
	CamConfigSendFIFOCmds();
}


