;    Copyright (C) 2004    John Orlando
;
;   AVRcam: a small real-time image processing engine.

;    This program is free software; you can redistribute it and/or
;    modify it under the terms of the GNU General Public
;    License as published by the Free Software Foundation; either
;    version 2 of the License, or (at your option) any later version.

;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;    General Public License for more details.

;    You should have received a copy of the GNU General Public
;    License along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

;   For more information on the AVRcam, please contact:

;   john@jrobot.net

;   or go to www.jrobot.net for more details regarding the system.
;**********************************************************************
;       Module Name: CanInterfaceAsm.S
;       Module Date: 04/14/2004
;       Module Auth: John Orlando
;
;       Description: This module provides the low-level interface
;       to the OV6620 camera hardware.  It is responsible for
;   	acquiring each pixel block (R,G,B), performing the mapping
;       into an actual color (orange, purple, etc), run-length
;       encoding the data, and storing the info off to the appropriate
;       line buffer.  This routine is synchronized with the pixel data
;       so that no polling of the camera data needs to be done (the
;       OV6620 is clocked off of the same crystal source as the mega8,
;       thus providing inherent synchronization between the two).
;
;       Revision History:
;       Date        Rel Ver.    Notes
;       4/10/2004      0.1     Module created
;       6/30/2004      1.0     Initial release for Circuit Cellar
;                              contest.
;       1/16/2005      1.4     Fixed issue with the TCCR1B register
;                              where PCLK was getting routed to the
;                              timer1 even when it wasn't needed.
;                              This caused excessive counter overflow
;                              interrupts, and caused problems.  Now,
;                              the "PCLK" pipe feeds timer1 when needed,
;                              and is turned off when it isn't needed.

#include "ATMega8.inc"
#include "EventCodes.h"

#define HREF_INTERRUPT_ENABLE_MASK            0x80
#define HREF_INTERRUPT_DISABLE_MASK           0x7F
#define ENABLE_PCLK_TIMER1_OVERFLOW_BITMASK   0x04
#define DISABLE_PCLK_TIMER1_OVERFLOW_BITMASK  0xFB
#define G_PORT						                    PortCPins
#define RB_PORT						                    PortBPins
#define PIXEL_RUN_START_INITIAL               0x50     	; This value causes our pixel counter (TCNT1)
                                                        ; to overflow after 176 (horizontal) pixels

#define RED_MEM_OFFSET				                0x00
#define GREEN_MEM_OFFSET			                0x10
#define BLUE_MEM_OFFSET				                0x20

; A pixelBlock is defined as a contiguous group of 4 pixels that are combined
; together to form a specific color.  Typically, this is formed by sampling a
; a green value, followed by a red and blue value (since we are dealing
; with Bayer color data).  We could optionally sample a second green with
; the red and average the greens, because the eye is more sensitive to
; green, but for speed we don't do this.  These three values (RGB) are then
; used as indices into the color membership lookup table (memLookup) to
; determine which color the pixelBlock maps into.  The memLookup table is
; manually generated for now (though it will hopefully be modified over
; the serial interface eventually).
;
; Here is a pixel block:
; ...G  G  G  G...  (row x)
; ...B  R  B  R...  (row x+1)
;    |  |  |  |--this is skipped
;    |  |  |--this is skipped
;    |  |--this is sampled
;    |--this is sampled

; As pixel blocks are sampled, the red, green, and blue values are
; used to index into their respective color maps.  The color maps
; return values that can be logically ANDed together so that a
; particular RGB triplet will result in a single bit being set
; after the AND operation.  This single bit indicates which color
; the RGB triplet represents.  It is also possible for no bits to
; be set after the AND process, indicating that the RGB triplet
; does not map to any of the colors configured in the color map.
; This isn't quite as fast as a pure RGB lookup table, but
; it then again it doesn't require 2^12 (4-bits for each color
; channel) bytes to store the lookup table.  It takes just a few
; more cycles, and only requires 48 bytes of precious RAM (16
; per color channel, since our resolution on each color channel
; is only 4-bits).  Not bad....for more information, see:
; http://www.cs.cmu.edu/~trb/papers/wirevision00.pdf for more
; information on this color segmentation technique.

; One other note: this code does depend on the colorMap residing
; at a well-defined position in memory; specifically, it mus
; start at a 256-byte boundary so that the lowest byte in the
; map is set to 0x00.  Currently, the colorMap is forced to
; start at RAM location 0x300.  This could potentially be changed
; by the developer if needed, but offsets would have to be added
; in to the colorMap look-up code below to make it work.


; These are the registers that will be used throughout this
; module for acquiring each line of pixel data

#define PixelCount                R20
#define PixelRunStart             R21
#define LastColor                 R22
#define Temp1                     R23
#define Temp2                     R24
#define Color                     R23
#define GreenData                 R24
#define BlueData                  R25
#define ColorMapLow               R18
#define ColorMapHigh              R19
#define PrevLineBuffLow           R18
#define PrevLineBuffHigh          R19
#define CurrLineBuffLow           R16
#define CurrLineBuffHigh          R17

#define DebugPin                  PortDData, 6

                    NAME CamInterfaceAsm
                    COMMON INTVEC:CODE:ROOT(1)
                    ORG ExternalInt0Vector
                    EXTERN ExternalInterrupt0
                    RJMP ExternalInterrupt0
                    ORG ExternalInt1Vector
                    EXTERN ExternalInterrupt1
                    RJMP ExternalInterrupt1
                    ORG Timer0OverflowIntVector
                    EXTERN Timer0OverflowInterrupt
                    RJMP Timer0OverflowInterrupt
                    ORG Timer1OverflowIntVector
                    EXTERN Timer1OverflowInterrupt
                    RJMP Timer1OverflowInterrupt
                    ENDMOD

                    NAME CamInt

; These are the global assembly function names that are accessed via other
; C functions
                    RSEG CODE
                    PUBLIC CamIntAsmWaitForNewTrackingFrame
                    PUBLIC CamIntAsmWaitForNewTrackingLine
                    PUBLIC CamIntAsmWaitForNewDumpFrame
                    PUBLIC CamIntAsmAcquireDumpLine
                    PUBLIC CamIntAsmAcquireTrackingLine
                    PUBLIC ExternalInterrupt0
                    PUBLIC ExternalInterrupt1
                    PUBLIC Timer0OverflowInterrupt
                    PUBLIC Timer1OverflowInterrupt
                    //EXTERN CurrentLineBuffer
                    //EXTERN ColorMap
                    EXTERN FastEventBitmask
                    EXTERN CoMaHigh
                    EXTERN CoMaLow
                    EXTERN EEPROMInit
                    EXTERN LineCount
                    EXTERN CurrentLineBuffer
                    EXTERN PreviousLineBuffer
                    EXTERN UARTTransmitByte

;*****************************************************************
;   	Function Name: CamIntAsm_waitForNewTrackingFrame
;       Function Description: This function is responsible for
;       going to sleep until a new frame begins (indicated by
;    	VSYNC transitioning from low to high.  This will wake
;       the "VSYNC sleep" up and allow it to continue with
;       the acquireLine function, where the system waits for
;       an "HREF sleep" that we use to synchronize with the
;       data.
;       Inputs:  r25 - MSB of currentLineBuffer
;                r24 - LSB of currentLineBuffer
;				 r23 - MSB of colorMap
; 				 r22 - LSB of colorMap
;       Outputs: none
;       NOTES: This function doesn't really return...it sorta just
;       floats into the acquireLine function after the "VSYNC sleep"
;       is awoken, then begins processing the line data.  Once
;		176 pixels are sampled (and the counter overflows), then
;		an interrupt will occur, the 'T' bit in the SREG will be
;		set, and the function will return.
;*****************************************************************
CamIntAsmWaitForNewTrackingLine
TRWaitForHrefLow    SBIC PortDPins, 4
                    RJMP TRWaitForHrefLow
                    RJMP CamIntAsmAcquireTrackingLine
                    
CamIntAsmWaitForNewTrackingFrame
WaitVSyncHigh:      SBIS PortDPins, 2                 ;   VSync
                    RJMP WaitVSyncHigh
;*****************************************************************
; REMEMBER...everything from here on out is critically timed to be
; synchronized with the flow of pixel data from the camera...
;    :-)      VSYNC
;    HREF 1.220 mS.   
;    .
;*****************************************************************

CamIntAsmAcquireTrackingLine
                    BRTS CleanUp
                    MOVW R1:R0, XHigh:XLow
                    MOVW R3:R2, YHigh:YLow
                    LDI Temp1, TimerRisingExternalPinEdge;ORI Temp1, TimerRisingExternalPinEdge
                    OUT Timer1ControlB, Temp1
                    ; The line is about to start...
                    LDI PixelCount, EEPROMInit  //0                           ;Initilaize the RLE stats...
                    LDI PixelRunStart, PIXEL_RUN_START_INITIAL  ;Remember, we always calculate
                                                                ;the pixel run length as
                                                                ;Timer1ValueLow - PixelRunStart
                    LDI LastColor, 0x00                         ;Clear out the last color before we start
                    ;MOV XHigh, CurrLineBuffHigh                 ;Load the pointer to the current line
                    ;MOV XLow, CurrLineBuffLow                   ;buffer into the X pointer regs (        )
                    MOVW XHigh:XLow, CurrLineBuffHigh:CurrLineBuffLow
                    MOVW CurrLineBuffHigh:CurrLineBuffLow, BlueData:Temp2 ;     CurrLineBuff     , 
                                                                 ;         ,  
                                                                 ;    .        
                                                                 ;    IAR  R24  R25
                    ;MOV ZHigh, ColorMapHigh                     ;Load the pointers to the membership
                    ;MOV ZLow, ColorMapLow                       ;lookup tables (ZLow and YLow will be overwritten
                    MOVW ZHigh:ZLow, ColorMapHigh:ColorMapLow
                    ;MOV YHigh, ColorMapHigh                     ;as soon as we start reading data) to Z and Y
                    IN Temp1, TimerInterruptMask                ;enable Timer1 to start counting      ??? ???
                    ORI Temp1, Timer1OverflowInterruptEnable    ;external PCLK pulses and interrupt on
                    OUT TimerInterruptMask, Temp1               ;overflow
                    LDI Temp1, PIXEL_RUN_START_INITIAL          ;Set up the Timer1 to overflow (and
                    LDI Temp2, 0xFF                             ;interrupts) after 176 pixels
                    OUT Timer1ValueHigh, Temp2
                    OUT Timer1ValueLow, Temp1
                    ;MOV YLow, ColorMapLow
                    MOVW YHigh:YLow, ColorMapHigh:ColorMapLow
                    IN Temp1, GeneralInterruptControl           ;Enable the HREF interrupt...remember? we
                                                                ;only use this interrupt to syncronize
                    ORI Temp1, HREF_INTERRUPT_ENABLE_MASK       ;the beginning of the line
                    OUT GeneralInterruptControl, Temp1          ;      4 

;*******************************************************************************************
;   Track Frame handler
;*******************************************************************************************

TrackFrame          SBI DebugPin
                    Sleep                                       ;...and we wait... (    c  HREF,  PCLK  )
                    ; Returning from the interrupt/sleep wakeup will consume
                    ; 14 clock cycles (7 to wakeup from idle sleep, 3 to vector, and 4 to return)
                    ; Disable the HREF interrupt
                    CBI DebugPin  ;2
                    IN Temp1, GeneralInterruptControl ;1
                    ANDI Temp1, HREF_INTERRUPT_DISABLE_MASK ;1
                    OUT GeneralInterruptControl, Temp1  ;1

                    ; A couple of NOPs are needed here to sync up the pixel data...the number (2)
                    ; of NOPs was determined emperically by trial and error.
                    NOP
                    NOP
                    ; Clock Cycle Count
AcquirePixelBlock   IN ZLow, RB_PORT                            ;sample the red value (PortBPins)     (1)
                    IN YLow, G_PORT                             ;sample the green value (PortCPins)   (1)
                    ANDI YLow, 0x0F                             ;clear the high nibble (of green)     (1)
                    LDD Color, Z+RED_MEM_OFFSET                 ;lookup the red membership            (2)
                    IN ZLow, RB_PORT                            ;sample the blue value (PortBPins)    (1)
                    LDD GreenData, Y+GREEN_MEM_OFFSET           ;lookup the green membership          (2)
                    LDD BlueData, Z+BLUE_MEM_OFFSET             ;lookup the blue membership           (2)
                    AND Color, GreenData                        ;mask memberships together            (1)
                    AND Color, BlueData                         ;to produce the final color           (1)
                    BRTS CleanUpTrackingLine                    ;if some interrupt routine has        (1...not set)
                                                                ;come in and set our T flag in
                                                                ;StatusRegister? then we need to hop out
                                                                ;and blow away this frames data (common cleanup)
                    CP Color, LastColor                         ;check to see if the run continues    (1)
                    BREQ AcquirePixelBlock                      ;                                     (2...equal)
                                                                ;                 _______________
                                                                ;                 16 clock cycles
                                                                ; (16 clock cycles = 1uS = 1 PixelBlock time)
                    ;Toggle the debug line to indicate a color change
                    SBI DebugPin
                    NOP
                    CBI DebugPin
                    MOV Temp2, PixelRunStart                    ;Get the count value of the
                                                                ;current pixel run
                    IN PixelCount, Timer1ValueLow               ;get the current Timer1ValueLow
                    MOV PixelRunStart, PixelCount               ;reload PixelRunStart for the next run
                    SUB PixelCount, Temp2                       ;PixelCount = Timer1ValueLow - PixelRunStart
                    ST X+, LastColor                            ;record the color run in the current line buffer
                    ST X+, PixelCount                           ;with it's length
                    MOV LastColor, Color                        ;set LastColor so we can figure out when it changes
                    NOP                                         ;waste one more cycle for a total of 16
                    RJMP AcquirePixelBlock

; CleanUpTrackingLine is used to write the last run length block off to the CurrentLineBuffer so
; that all 176 pixels in the line are accounted for.
CleanUpTrackingLine 
                    LDI PixelCount, 0xFF                        ;the length of the last run is ALWAYS 0xFF minus the last
                    SUB PixelCount, PixelRunStart               ;PixelRunStart
                    INC PixelCount                              ;increment PixelCount since we actually need to account
                                                                ;for the overflow of Timer1
                    ST X+, Color                                ;record the color run in the current line buffer
                    ST X, PixelCount
                    RJMP CleanUp

CleanUpDumpLine
		; NOTE: If serial data is received, to interrupt the tracking of a line, we'll
		; get a EV_SERIAL_DATA_RECEIVED event, and the T bit set so we will end the
		; line's processing...however, the PCLK will keep on ticking for the rest of
		; the frame/line, which will cause the TCNT to eventually overflow and
		; interrupt us, generating a EV_ACQUIRE_LINE_COMPLETE event.  We don't want
		; this, so we need to actually turn off the PCLK counting each time we exit
		; this loop, and only turn it on when we begin acquiring lines....
                    ; NOT NEEDED FOR NOW...
                    IN Temp1, TimerInterruptMask			; disable TIMER1 to stop counting
                    ANDI Temp1, DISABLE_PCLK_TIMER1_OVERFLOW_BITMASK 	; external PCLK pulses
                    OUT TimerInterruptMask, Temp1
CleanUp
                    ; Disable the external clocking of the Timer1 counter
                    IN Temp1, Timer1ControlB
                    ANDI Temp1, 0xF8
                    OUT Timer1ControlB, Temp1
                    ;Toggle the debug line to indicate the line is complete
                    ;SBI DebugPin
                    ;CBI DebugPin
                    MOVW BlueData:Temp2, CurrLineBuffHigh:CurrLineBuffLow
                    MOVW YHigh:YLow, R3:R2
                    MOVW XHigh:XLow, R1:R0
                    CLT                                         ;clear out the T bit since we have detected
                                                                ;the interruption and are exiting to handle it
                    CBI PortDData, 7
Exit                RET
;*****************************************************************
;   	Function Name: CamIntAsm_waitForNewDumpFrame
;       Function Description: This function is responsible for
;       going to sleep until a new frame begins (indicated by
;    	VSYNC transitioning from low to high.  This will wake
;       the "VSYNC sleep" up and allow it to continue with
;       acquiring a line of pixel data to dump out to the UI.
;       Inputs:  r25 - MSB of currentLineBuffer
;                r24 - LSB of currentLineBuffer
;				 Temp1 - MSB of prevLineBuffer
;				 r22 - LSB of prevLineBuffer
;       Outputs: none
;       NOTES: This function doesn't really return...it sorta just
;       floats into the acquireDumpLine function after the "VSYNC sleep"
;       is awoken.
;*****************************************************************
CamIntAsmWaitForNewDumpFrame
                    MOVW R1:R0, XHigh:XLow
                    MOVW R3:R2, YHigh:YLow
DFWaitForVSyncHigh  SBIS PortDPins, 2
                    RJMP DFWaitForVSyncHigh
DFWaitForVSyncLow   SBIC PortDPins, 2
                    RJMP DFWaitForVSyncLow
                    LDS Temp1, LineCount
                    LSL Temp1 
DFLineCountCycle    BREQ CamIntAsmAcquireDumpLine
DFWaitForHrefHigh   SBIS PortDPins, 4
                    RJMP DFWaitForHrefHigh
                    DEC Temp1
DFWaitForHrefLow    SBIC PortDPins, 4
                    RJMP DFWaitForHrefLow
                    RJMP DFLineCountCycle

;*****************************************************************
; REMEMBER...everything from here on out is critically timed to be
; synchronized with the flow of pixel data from the camera...
;*****************************************************************
CamIntAsmAcquireDumpLine
                    BRTS CleanUp 
                    MOVW XHigh:XLow, CurrLineBuffHigh:CurrLineBuffLow ;Load the pointer to the current line
                                                                ;buffer into the X pointer regs
                    MOVW CurrLineBuffHigh:CurrLineBuffLow, BlueData:Temp2 ;-     Temp2,
                                                                ;     DumpFrame  TrackingFrame  
                                                                ;  TrackingFrame   BlueData, 
                                                                ;     
                    MOVW YHigh:YLow, PrevLineBuffHigh:PrevLineBuffLow ;Load the pointer to the previous line
                                                                ;buffer into the Y pointer regs
                    LDI Temp1, PIXEL_RUN_START_INITIAL          ;set up the Timer1Value to overflow (and
                    LDI Temp2, 0xFF                             ;interrupts) after 176 pixels
                    OUT Timer1ValueHigh, Temp2
                    OUT Timer1ValueLow, Temp1
                    IN Temp1, Timer1ControlB                    ;Enable the PCLK line to actually
                    ORI Temp1, TimerRisingExternalPinEdge       ;feed Timer1
                    OUT Timer1ControlB, Temp1
                    ;NOP
                    IN Temp1, TimerInterruptMask                ;enable Timer1 to start counting
                    ORI Temp1, Timer1OverflowInterruptEnable    ;external PCLK pulses and interrupt on
                    OUT TimerInterruptMask, Temp1               ;overflow
                    IN Temp1, GeneralInterruptControl           ;enable the HREF interrupt...remember, we
                                                                ;only use this interrupt to synchronize
                                                                ;the beginning of the line
                    ORI Temp1, HREF_INTERRUPT_ENABLE_MASK
                    OUT GeneralInterruptControl, Temp1
                    
;*******************************************************************************************
;   Dump Frame handler
;*******************************************************************************************
DumpFrame           SBI DebugPin
                    SLEEP                                       ;...and we wait...
                    CBI DebugPin
                    IN Temp1, GeneralInterruptControl           ;Disable the HREF interrupt
                    ANDI Temp1, HREF_INTERRUPT_DISABLE_MASK     ;so we don't get interrupted
                    OUT GeneralInterruptControl, Temp1          ;while dumping the line
                    NOP                                         ;Remember...if we ever remove the "CBI" instruction above,
                                                                ;we need to add two more NOP's to cover this
; Ok...the following loop needs to run in 8 clock cycles, so we can get every
; pixel in the line...this shouldn't be a problem, since the PCLK timing was
; reduced by a factor of 2 whenever we go to dump a line (this is to give us
; enough time to do the sampling and storing of the pixel data).  In addition,
; it is assumed that we will have to do some minor processing on the data right
; before we send it out, like mask off the top 4-bits of each, and then pack both
; low nibbles into a single byte for transmission...we just don't have time to
; do that here (only 8 instruction cycles :-)  )
SampleDumpPixel     IN Temp1, G_PORT                            ;sample the G value                     (1)
                    IN Temp2, RB_PORT                           ;sample the R/B value                   (1)
                    ST X+, Temp1                                ;store to the CurrLineBuff and inc ptrs (2)
                    ST Y+, Temp2                                ;store to the PrevLineBuff and inc ptrs (2)
                    BRTC SampleDumpPixel                        ;loop back unless flag is set           (2...if not set)
                                                                ;                     _________________
                                                                ;                     8 cycles normally
                    ;if we make it here, it means the T flag is set, and we must have been interrupted
                    ;so we need to exit (what if we were interrupted for serial? should we disable it?)
                    RJMP CleanUpDumpLine

;***********************************************************
;	Function Name: <interrupt handler for External Interrupt0>
;	Function Description: This function is responsible
;	for handling a rising edge on the Ext Interrupt 0.  This
;	routine simply returns, since we just want to wake up
;	whenever the VSYNC transitions (meaning the start of a new
;	frame).
;	Inputs:  none
;	Outputs: none
;***********************************************************
ExternalInterrupt0
                    ; This will wake us up when VSYNC transitions high...we just want to return
                    RETI

;***********************************************************
;	Function Name: <interrupt handler for External Interrupt1>
;	Function Description: This function is responsible
;	for handling a falling edge on the Ext Interrupt 1.  This
;	routine simply returns, since we just want to wake up
;	whenever the HREF transitions (meaning the pixels
;	are starting after VSYNC transitioned, and we need to
; 	start acquiring the pixel blocks
;	Inputs:  none
;	Outputs: none
;***********************************************************
ExternalInterrupt1
                    ; This will wake us up when HREF transitions high...we just want to return
                    RETI

;***********************************************************
;	Function Name: <interrupt handler for Timer0 overflow>
;	Function Description: This function is responsible
;	for handling the Timer0 overflow (hooked up to indicate
;	when we have reached the number of HREFs required in a
;	single frame).  We set the T flag in the SREG to
;	indicate to the _acquirePixelBlock routine that it needs
;	to exit, and then set the appropriate action to take in
;	the eventList of the Executive module.
;	Inputs:  none
;	Outputs: none
;   Note: Originally, the HREF pulses were also going to
;   be counted by a hardware counter, but it didn't end up
;   being necessary
;***********************************************************
Timer0OverflowInterrupt
;		                set				; set the T bit in SREG
;		                lds		tmp1,eventBitmask
;		                ori		tmp1,EV_ACQUIRE_FRAME_COMPLETE
;		                sts		eventBitmask,tmp1
                                ;CBI DebugPin
                                ;SBI DebugPin
		                reti

;***********************************************************
;	Function Name: <interrupt handler for Timer1 overflow>
;	Function Description: This function is responsible
;	for handling the Timer1 overflow (hooked up to indicate
;	when we have reached the end of a line of pixel data,
;	since PCLK is hooked up to overflow TCNT1 after 176
;	pixels).  This routine generates an acquire line complete
;	event in the fastEventBitmask, which is streamlined for
;	efficiency reasons.
;***********************************************************
Timer1OverflowInterrupt
                    LDS		Temp1, FastEventBitmask   		  ; set a flag indicating
                    ORI		Temp1,FEV_ACQUIRE_LINE_COMPLETE	; a line is complete
                    STS		FastEventBitmask,Temp1
                    SET		                                ; set the T bit in SREG
                    ;SBI		DebugPin                  ; For testing...
                    ;SBI		DebugPin                  ; For testing...
                    RETI

                    END
