/*
    Copyright (C) 2004    John Orlando

   AVRcam: a small real-time image processing engine.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   For more information on the AVRcam, please contact:

   john@jrobot.net

   or go to www.jrobot.net for more details regarding the system.
*/
/***********************************************************
	Module Name: I2CInterface.c
	Module Date: 4/10/2004
	Module Auth: John Orlando

	Description: This module is responsible for providing a
	low-level interface to the I2C hardware resident on the
	mega8 processor (also known as the Two-Wire Interface,
	or TWI).  The interface is needed to configure the
	needed registers in the OV6620 camera.  This interface
	is interrupt-driven based on the events that should
	occur upon successful writing of an I2C register.

    Revision History:
    Date        Rel Ver.    Notes
    4/10/2004      0.1     Module created
    6/30/2004      1.0     Initial release for Circuit Cellar
                           contest.

***********************************************************/

/*	Includes */
#include "I2CStatusStates.h"

/*  Structures and Typedefs */
typedef struct
{
	unsigned char ConfigReg;
	unsigned char Data;
} I2CCommand;       //i2cCmd_t;

/*  Defines   */
/*  The I2C interface will run at 100 KHz (the OV6620 can
	support up to 400 KHz, but this hasn't been tested) */


/*  External Function Prototypes */

/*  Local Variables */

/* These variables are used as storage space for the current
   I2C command being sent over the interface.  They need to
   be volatile since they are dealt with an the TWI ISR */
volatile unsigned char I2CAddress;
volatile unsigned char *I2CData;
volatile unsigned char I2CDirection;
volatile unsigned char I2CBytes;
volatile unsigned char I2CStatus;
volatile unsigned char I2CRetryCounter;

/* 	Local Structures and Typedefs */

/*  Extern Variables */

/*  Definitions */
/* Bit definitions for the tw_status register */
#define MAX_TWI_RETRIES 2

/***********************************************************
	Function Name: I2CInt_init
	Function Description: This function is responsible
	for setting up the registers needed for the TWI
	interface

	Inputs:  none
	Outputs: none
***********************************************************/
void I2CInit(void)
{
  TWIStatus=0x00;
  /* init the speed of the I2C interface, running at
    100 Kbps */
  TWIBitRate=(CPUFrequency/I2C_BIT_RATE-16)/2;
  TWIControl=TWIInterruptFlag|TWIEnableAcknowledge|TWIEnable|TWIInterruptEnable;
}

/***********************************************************
	Function Name: I2CInt_writeData
	Function Description: This function is responsible for
	initiating the process of writing a sequence of bytes
	an I2C slave address.  This function will try to write
	the data three times before giving up.
	Inputs: address: the address of the I2C slave device
			data: a pointer to the data to be written
				  to the slave...for camera interfacing,
				  the data follows a <register #><data>
				  format
			bytes: the number of bytes to write
	Outputs: none
***********************************************************/
void I2CWriteData(unsigned char address, unsigned char *data, unsigned char bytes)
{
  while (I2CStatus&BUSY);       /* Bus is busy wait (or exit with error code) */            //   
  while (TWIControl&TWIStopCondition);      //  
  /* copy the needed data and state info to our local I2C command structure */
  I2CAddress=address;
  I2CData=data;
  I2CBytes=bytes;
  I2CDirection=I2C_WRITE;
  I2CRetryCounter=0;
  /* Generate start condition, the remainder of the transfer is interrupt driven and
	   will be performed in the background */
  TWIControl=TWIInterruptFlag|TWIEnableAcknowledge|TWIStartCondition|TWIEnable|TWIInterruptEnable;
  I2CStatus|=BUSY;
}

/***********************************************************
	Function Name: I2CInt_readData
	Function Description: This funcion is responsible for
	reading the specified number of bytes from a slave
	device.
	Inputs:  address: the slave address to read from
			 data: a pointer to where the data will be stored
			 bytes: the number of bytes to read
	Outputs: none
***********************************************************/
void I2CReadData(unsigned char address, unsigned char *data, unsigned char bytes)
{
    /* Bus is busy wait (or exit with error code) */
	while(I2CStatus & BUSY);

	I2CAddress = address;
	I2CData = data;
	I2CBytes = bytes;
	I2CDirection = I2C_READ;

	I2CRetryCounter = 0;

	/* Generate start condition, the remainder of the transfer is interrupt driven and
	   will be performed in the background */
	TWIControl = TWIInterruptFlag|TWIStartCondition|TWIEnable|TWIEnableAcknowledge|TWIInterruptEnable;

	I2CStatus |= BUSY;
}

/***********************************************************
	Function Name: I2CInt_isI2cBusy
	Function Description: This funcion is responsible for
	indicating if the I2C bus is currently busy to external
	modules.
	device.
	Inputs:  none
	Outputs: bool_t - indicating if bus is busy
***********************************************************/
unsigned char IsI2CBusy(void)
{
	unsigned char RetVal = FALSE;
	if ( (I2CStatus & BUSY) != 0)
	{
		RetVal = TRUE;
	}

	return(RetVal);
}

/***********************************************************
	Function Name: <interrupt handler for I2C>
	Function Description: This function is responsible for
	implementing the control logic needed to perform a
	read or write operation with an I2C slave.
	Inputs:  none
	Outputs: none
***********************************************************/
#pragma vector=TwoWireSerialIntVector
__interrupt void I2CInterrupt(void)
{
	unsigned char TWI_status = TWIStatus & 0xF8;   // grab just the status bits 

    // the entire I2C handler is state-based...determine
    //what needs to be done based on TWI_status 
	switch(TWI_status)
    {
        case TWI_START:									// Start condition 
        case TWI_REP_START:								// Repeated start condition 
            if(I2CRetryCounter > MAX_TWI_RETRIES)
            {
                // generate stop condition if we've reached our retry limit 
                TWIControl|=TWIInterruptFlag|TWIStopCondition;
                I2CStatus &= ~BUSY;
                return;
            }
            // indicate read or write 
            TWIData=(I2CAddress<<1)+I2CDirection;
            // TWSTA must be cleared...also clears TWINT 
            TWIControl &= ~TWIStartCondition;
            break;

        case TWI_MT_SLA_ACK:							// Slave acknowledged address, 
            I2CRetryCounter = 0;
            // tx the data, and increment the data pointer 
            TWIData = *I2CData;
            I2CData++;

            // clear the int to continue 
            TWIControl |= TWIInterruptFlag;
            break;

        case TWI_MT_SLA_NACK:							// Slave didn't acknowledge address, 
        case TWI_MR_SLA_NACK:
            I2CRetryCounter++;

            // retry...
            TWIControl |= TWIInterruptFlag|TWIStopCondition|TWIStartCondition;
            break;

        case TWI_MT_DATA_ACK:							// Slave Acknowledged data, 
            if(--I2CBytes > 0)
            {
                //more data to send, so send it 
                TWIData = *I2CData;
                I2CData++;
                TWIControl |= TWIInterruptFlag;
            }
            else
            {
                // generate the stop condition if needed 
                TWIControl |= TWIStopCondition|TWIInterruptFlag;
                I2CStatus &= ~BUSY;
            }
            break;

        case TWI_MT_DATA_NACK:							// Slave didn't acknowledge data 
            // send the stop condition 
            TWIControl |= TWIStopCondition|TWIInterruptFlag;
            I2CStatus &= ~BUSY;
            break;

        case TWI_MR_SLA_ACK:                             // Slave acknowledged address 
            if(--I2CBytes > 0)
            {
                // if there is more than one byte to read, acknowledge 
                TWIControl |= TWIEnableAcknowledge|TWIInterruptFlag;
            }
			else
            {
                // no acknowledge 
                TWIControl |= TWIInterruptFlag;
            }
            break;

        case TWI_MR_DATA_ACK: 							// Master acknowledged data 

            // grab the received data 
            *I2CData = TWIData;
            I2CData++;
            if(--I2CBytes > 0)
            {
                // get the next data byte and ack 
                TWIControl |= TWIEnableAcknowledge|TWIInterruptFlag;
            }
            else
            {
                // clear out the enable acknowledge bit 
                TWIControl &= ~TWIEnableAcknowledge;
            }
            break;

        case TWI_MR_DATA_NACK:						// Master didn't acknowledge data -> end of read process 
            // read data, and generate the stop condition 
            *I2CData = TWIData;
            TWIControl |= TWIStopCondition|TWIInterruptFlag;
            I2CStatus &= ~BUSY;
            break;
	}
}
