/*
    Copyright (C) 2004    John Orlando

   AVRcam: a small real-time image processing engine.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   For more information on the AVRcam, please contact:

   john@jrobot.net

   or go to www.jrobot.net for more details regarding the system.
*/
/***********************************************************
	Module Name: Uart.h
	Module Date: 04/10/2004
	Module Auth: John Orlando

	Description: This module is responsible for providing an
	interface to the UART hardware available on the mega8.
	This interface is an interrupt-driven interface.

    Revision History:
    Date        Rel Ver.    Notes
    4/10/2004      0.1     Module created
    6/30/2004      1.0     Initial release for Circuit Cellar
                           contest.
    11/15/2004     1.2     Updated UART baud rate regs so that
                           it runs at 115.2 kbps when the input
                           crystal is at 17.7 MHz (which is the
                           speed of the OV6620's crystal).
    1/16/2005      1.4     Moved the serial received ISR to
                           this file, instead of having it
                           in its own UartInterfaceAsm.S file
                           written in assembly.
***********************************************************/

/*	Includes */

/*  Defines */
#define COMMAND_RX_FIFO_SIZE 32
#define COMMAND_RX_FIFO_MASK COMMAND_RX_FIFO_SIZE-1

#define COMMAND_TX_FIFO_SIZE 64
#define COMMAND_TX_FIFO_MASK COMMAND_TX_FIFO_SIZE-1

/*  Local Variables */

/* 	Local Structures and Typedefs */

/*  Extern Variables */
unsigned char CommandRxFifo[COMMAND_RX_FIFO_SIZE]; //UIMgr_rxFifo[UI_MGR_RX_FIFO_SIZE];
unsigned char CommandRxFifoHead=0;
unsigned char CommandRxFifoTail=0;

unsigned char CommandTxFifo[COMMAND_TX_FIFO_SIZE]; //UIMgr_txFifo[UI_MGR_TX_FIFO_SIZE];
unsigned char CommandTxFifoHead=0;
unsigned char CommandTxFifoTail=0;

/*  Definitions */

/***********************************************************
	Function Name: UartInt_init
	Function Description: This function is responsible for
	initializing the UART interface on the mega8.  This
	interface is set to communicate at 115.2 Kbps, with an
	8N1 protocol.
	Inputs:  none
	Outputs: none
***********************************************************/
void UartInit(void)
{
  /* set up the baud rate registers so the UART will operate
  at 115.2 Kbps */
  UARTBaudRateHigh = 0x00;

#ifdef NO_CRYSTAL
  UARTBaudRateLow = 18;  /* 18 for double clocking at 115.2 kbps */
#else
  UARTBaudRateLow = 0x08;  /* for 16 MHz crystal at 115.2 kbps */
#endif
  
  /* enable the tx and rx capabilities of the UART...as well
  as the receive complete interrupt */
  UARTControl=UARTTransmitterEnable|UARTReceiverEnable|UARTReceiveInterruptEnable;
  /* set up the control registers so the UART works at 8N1 */
  UARTSettings=UART8BitCharacters|UARTControlRegisterSelect;

#ifdef NO_CRYSTAL
    /* set the baud rate to use the double-speed */
  UARTStatus=UARTDoubleSpeed;
#endif
}

/***********************************************************
  Function Name: UartInt_txByte
  Function Description: This function is responsible for
  transmitting a single byte on the uart.
  Inputs:  txByte - the byte to send
  Outputs: none
  NOTES: When the TX UDRE (data register empty) is set, there
  is puposefully no interrupt...thus, to send a string of
  data out, the calling routine needs to hold up the entire
  application while this takes place (or just send one
  byte at a time at strtegically timed intervals, like
  the stats data is sent out :-)
***********************************************************/
void UARTTransmitByte(unsigned char Byte)
{
  /* Wait for empty transmit buffer */
  while ( !( UARTStatus & UARTDataRegisterEmpty) );
  /* Put data into buffer, sends the data */
  UARTData = Byte;
}

/***********************************************************
  Function Name: SIG_UART_RECV ISR
  Function Description: This function is responsible for
  handling the interrupt caused when a data byte is
  received by the UART.
  Inputs:  none
  Outputs: none
  NOTES: This function was originally written in assembly,
  but moved over to C when the setting of the "T" bit at
  the end of the routine was no longer necessary (this
  theoretically allowed the AVRcam to respond to serial
  bytes in the middle of tracking or dumping a frame.
  But it wasn't really needed, and understanding the C
  is easier  :-)
***********************************************************/
/*#pragma vector=UARTReceiveIntVector
__interrupt void UARTReceiveInterrupt(void)
{
  CommandRxFifo[CommandRxFifoHead++] = UARTData;
  CommandRxFifoHead &= COMMAND_RX_FIFO_MASK;
  PlaceEvent(EV_SERIAL_DATA_RECEIVED);
}*/

