/*
    Copyright (C) 2004    John Orlando

   AVRcam: a small real-time image processing engine.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   For more information on the AVRcam, please contact:

   john@jrobot.net

   or go to www.jrobot.net for more details regarding the system.
*/
/***********************************************************
	Module Name: UIMgr.c
	Module Date: 04/10/2004
	Module Auth: John Orlando

	Description: This module is responsible for providing
	the processing to manage the user interface of the
	system.  This user interface is provided via the UART.
	This module handles the incoming serial commands, and
	performs the needed functionality.  It is then
	responsible for generating any needed response to
	the external entity.

    Revision History:
    Date        Rel Ver.    Notes
    4/10/2004      0.1     Module created
    6/30/2004      1.0     Initial release for Circuit Cellar
                           contest.
    11/15/2004     1.2     Updated version string to 1.2.
    1/16/2005      1.4     Added code to write the colorMap
                           to EEPROM one byte at a time,
                           ensuring that the EEPROM is only
                           written when the data is different
                           than the current value (thus saving
                           EEPROM writes).  Updated version
                           string to 1.4.
***********************************************************/

/*	Includes */

/* Defines */
#define UI_MGR_RX_FIFO_SIZE 32 
#define UI_MGR_RX_FIFO_MASK UI_MGR_RX_FIFO_SIZE-1 

#define UI_MGR_TX_FIFO_SIZE 64
#define UI_MGR_TX_FIFO_MASK UI_MGR_TX_FIFO_SIZE-1

#define LENGTH_OF_UI_CMD 		2
#define UI_MGR_MAX_CMD_LENGTH 	32
#define MAX_TOKEN_LENGTH		3   /* needs to hold 0 - 255 in ASCII format */
#define MAX_TOKEN_COUNT			64  /* for setting the color maps */

/* 	Local Structures and Typedefs */

typedef enum
{
	GetVersionCmd,
	PingCmd,
	SetCameraRegsCmd,
	DumpFrameCmd,
	EnableTrackingCmd,
	DisableTrackingCmd,
	SetColorMapCmd,
	ResetCameraCmd,
	NoCmd,
	InvalidCmd
} CommandEnum; //UIMgr_Cmd_t;

typedef enum
{
	SetRed,
	SetGreen,
	SetBlue
} SetColorEnum; //setColorState_t;


/*  Local Variables */
static unsigned char CharCount = 0;
static unsigned char CharIndex = 0;
static unsigned char ASCIITokenBuffer[MAX_TOKEN_LENGTH+1]; /* +1 to ensure NULL at end */
static unsigned char TokenCount = 0;
static unsigned char TokenBuffer[MAX_TOKEN_COUNT];
static CommandEnum ReceivedCmd = NoCmd;
const __flash unsigned char AVRCamVersion[] = "AVRCam v1.4\r";

/*  Local Function Declaration */
//static unsigned char UIMgr_readRxFifo(void);
//static unsigned char UIMgr_readTxFifo(void);
//static unsigned char UIMgr_readRxFifo(void);
//static void UIMgr_sendNck(void);
//static void UIMgr_sendAck(void);
//static void UIMgr_convertTokenToCmd(void);
//static void UIMgr_convertTokenToValue(void);
//static void UIMgr_executeCmd(void);

/*  Extern Variables */

/*  Definitions */
#define IS_DATA_IN_TX_FIFO() (!(CommandTxFifoHead == CommandTxFifoTail))
#define IS_DATA_IN_RX_FIFO() (!(CommandRxFifoHead == CommandRxFifoTail))

/* MAX_EEPROM_WRITE_ATTEMPTS limits the number of writes that can be
done to a particular EEPROM cell, so that it can't possible just
write to the same cell over and over */
#define MAX_EEPROM_WRITE_ATTEMPTS 3

/***********************************************************
	Function Name: UIMgr_init
	Function Description: This function is responsible for
	initializing the UIMgr module.  It sets up the fifo
	used to hold incoming data, etc.
	Inputs:  none
	Outputs: none
***********************************************************/
void UIManagerInit(void)
{
	MemSet(ASCIITokenBuffer,0x00,MAX_TOKEN_LENGTH+1);
	MemSet(TokenBuffer,0x00,MAX_TOKEN_COUNT);
	MemSet(CommandTxFifo,0x00,COMMAND_TX_FIFO_SIZE);
	MemSet(CommandRxFifo,0x00,COMMAND_RX_FIFO_SIZE);
}

/***********************************************************
	Function Name: UIMgr_readTxFifo
	Function Description: This function is responsible for
	reading a single byte of data from the tx fifo, and
	updating the appropriate pointers.
	Inputs:  none
	Outputs: unsigned char-the data read
***********************************************************/
static unsigned char CommandReadTxFifo(void)
{
	unsigned char DataByte, TempTail;

	/* just return the current tail from the tx fifo */
	DisableInterrupts();
	DataByte = CommandTxFifo[CommandTxFifoTail];
	TempTail = (CommandTxFifoTail+1) & (COMMAND_TX_FIFO_MASK);
	CommandTxFifoTail = TempTail;
	EnableInterrupts();

	return(DataByte);
}

/***********************************************************
	Function Name: UIMgr_readRxFifo
	Function Description: This function is responsible for
	reading a single byte of data from the rx fifo, and
	updating the appropriate pointers.
	Inputs:  none
	Outputs: unsigned char-the data read
***********************************************************/
static unsigned char CommandReadRxFifo(void)
{
  unsigned char DataByte;

  /* just return the current tail from the rx fifo */
  DisableInterrupts();
  DataByte = CommandRxFifo[CommandRxFifoTail++];
  CommandRxFifoTail &= COMMAND_RX_FIFO_MASK;
  EnableInterrupts();
  return(DataByte);
}

/***********************************************************
	Function Name: UIMgr_writeTxFifo
	Function Description: This function is responsible for
	writing a single byte to the TxFifo and
	updating the appropriate pointers.
	Inputs:  data - the byte to write to the Fifo
	Outputs: none
***********************************************************/
void CommandWriteTxFifo(unsigned char Data)
{
  unsigned char TempHead;
  DisableInterrupts();
  CommandTxFifo[CommandTxFifoHead] = Data;
  /* now move the head up */
  TempHead = (CommandTxFifoHead + 1) & (UI_MGR_TX_FIFO_MASK);
  CommandTxFifoHead = TempHead;
  EnableInterrupts();
}

/***********************************************************
	Function Name: UIMgr_flushTxBuffer
	Function Description: This function is responsible for
	sending all data currently in the serial tx buffer
	to the user.
	Inputs:  none
	Outputs: none
***********************************************************/
void CommandFlushTxBuffer(void)
{
  while(CommandTxFifoHead != CommandTxFifoTail)
  {
    UARTTransmitByte(CommandReadTxFifo());
  }
}

/***********************************************************
	Function Name: UIMgr_transmitPendingData
	Function Description: This function is responsible for
	transmitting a single byte of data if data is waiting
	to be sent.  Otherwise, if nothing is waiting, the
	function just returns.
	Inputs:  none
	Outputs: none
***********************************************************/
void CommandTransmitPendingData(void)
{
  if (IS_DATA_IN_TX_FIFO() == TRUE)
  {
    /* data is waiting...send a single byte */
    UARTTransmitByte( CommandReadTxFifo() );
  }
}

/***********************************************************
	Function Name: UIMgr_convertTokenToCmd
	Function Description: This function is responsible for
	parsing a received 2-character command.  It will
	access the asciiTokenBuffer directly.
	Inputs:  none
	Outputs: none
***********************************************************/
static void ConvertTokenToCommand(void)
{
	if ( (ASCIITokenBuffer[0] == 'P') &&
		 (ASCIITokenBuffer[1] == 'G') )
	{
		/* we got a "ping" command...but we still need to see
		if we are going to get the \r */
		ReceivedCmd = PingCmd;
	}
	else if ( (ASCIITokenBuffer[0] == 'G') &&
			   (ASCIITokenBuffer[1] == 'V') )
	{
		/* we got the "get version" command */
		ReceivedCmd = GetVersionCmd;
	}
	else if ( (ASCIITokenBuffer[0] == 'D') &&
			   (ASCIITokenBuffer[1] == 'F') )
	{
		/* we should go into frame dump mode */
		ReceivedCmd = DumpFrameCmd;
	}
	else if ( (ASCIITokenBuffer[0] == 'C') &&
	           (ASCIITokenBuffer[1] == 'R') )
	{
		/* the user wants to set registers in the OV6620 */
		ReceivedCmd = SetCameraRegsCmd;
	}
	else if ( (ASCIITokenBuffer[0] == 'E') &&
			   (ASCIITokenBuffer[1] == 'T') )
	{
		/* the user wants to enable tracking */
		ReceivedCmd = EnableTrackingCmd;
	}
	else if ( (ASCIITokenBuffer[0] == 'S') &&
			   (ASCIITokenBuffer[1] == 'M') )
	{
		/* the user wants to set the color map */
		ReceivedCmd = SetColorMapCmd;
	}
	else if ( (ASCIITokenBuffer[0] == 'D') &&
			   (ASCIITokenBuffer[1] == 'T') )
	{
		ReceivedCmd = DisableTrackingCmd;
	}
	else if ( (ASCIITokenBuffer[0] == 'R') &&
			   (ASCIITokenBuffer[1] == 'S') )
	{
		ReceivedCmd = ResetCameraCmd;
	}
	else
	{
		/* don't recognize the cmd */
		ReceivedCmd = InvalidCmd;
	}
	MemSet(ASCIITokenBuffer,0x00,MAX_TOKEN_LENGTH);
	CharIndex = 0;
	CharCount = 0;
}

/***********************************************************
  Function Name: UIMgr_convertTokenToValue
  Function Description: This function is responsible for
  converting a received token to a hex value It will
  access the asciiTokenBuffer directly, and store the
  result in the appropriate token buffer.
  Inputs:  none
  Outputs: none
***********************************************************/
static void ConvertTokenToValue(void)
{
	unsigned int NewValue;

	NewValue = DecValue(ASCIITokenBuffer, 0);  //  3     (0-999)     255    
	if (NewValue > 255)     //     .    ,    unsigned char     
	{                       //Update:       .    ,      ,    
		/* the value is too large */  // ,   ,   ,   
		ReceivedCmd = InvalidCmd;
		TokenBuffer[TokenCount] = 0xFF;  /* to indicate an error */
	}
	else
	{
		/* copy the value into the tokenBuffer */
		TokenBuffer[TokenCount] = NewValue;
	}
	MemSet(ASCIITokenBuffer,0x00,MAX_TOKEN_LENGTH);
	CharIndex = 0;
	CharCount = 0;
}

/***********************************************************
	Function Name: UIMgr_sendNck
	Function Description: This function is responsible for
	queueing up an NCK to be sent to the user.
	Inputs:  none
	Outputs: none
***********************************************************/
static void CommandSendNck(void)
{
		CommandWriteTxFifo('N');    //.     
		CommandWriteTxFifo('C');
		CommandWriteTxFifo('K');
		CommandWriteTxFifo('\r');
}

/***********************************************************
	Function Name: UIMgr_sendAck
	Function Description: This function is responsible for
	queuing up an ACK to be sent to the user.
	Inputs:  none
	Outputs: none
***********************************************************/
static void CommandSendAck(void)
{
  CommandWriteTxFifo('A');
  CommandWriteTxFifo('C');
  CommandWriteTxFifo('K');
  CommandWriteTxFifo('\r');
}

/***********************************************************
	Function Name: UIMgr_txBuffer
	Function Description: This function is responsible for
	sending 'length' bytes out using the UartInterface
	module.
	Inputs:  pData -  a pointer to the data to send
	         length - the number of bytes to send
	Outputs: none
***********************************************************/
void PlaceDataInTxBuffer(unsigned char *pData, unsigned char Length)
{
	while(Length-- != 0)
	{
		UARTTransmitByte(*pData++);
	}
}

/***********************************************************
	Function Name: UIMgr_executeCmd
	Function Description: This function is responsible for
	executing whatever cmd is stored in the receivedCmd
	object.
	Inputs:  none
	Outputs: none
***********************************************************/
static void ExecuteCommand(void)
{
	unsigned char i,eepromData, NumWrites=0;
	const unsigned char __flash *pData;
    unsigned char EEPROMWriteSucceeded = FALSE;
#if	DEBUG_COLOR_MAP
	unsigned char ASCIIBuffer[5];
#endif

	if (ReceivedCmd == PingCmd)
	{
	}
	else if (ReceivedCmd == GetVersionCmd)
	{
		pData = AVRCamVersion;
		while(*pData != 0)
		{
			CommandWriteTxFifo(*pData++);
		}
	}
	else if (ReceivedCmd == ResetCameraCmd)
	{
		ResetCam(); //CamInt_resetCam();  //         ,     
	}
	else if (ReceivedCmd == DumpFrameCmd)
	{
		/* publish the event that will indicate that
		a request has come to dump a frame...this will
		be received by the FrameMgr, which will begin
		dumping the frame...a short delay is needed
		here to keep the Java demo app happy (sometimes
		it wouldn't be able to receive the serial data
		as quickly as AVRcam can provide it). */
		//Utility_delay(100);
		PlaceEvent(EV_DUMP_FRAME);
	}
	else if (ReceivedCmd == SetCameraRegsCmd)
	{
		/* we need to gather the tokens and
		build config cmds to be sent to the camera */
		for (i=1; i<TokenCount; i+=2)  /* starts at 1 since first token
											is the CR cmd */
		{
			CamConfigSetCamReg(TokenBuffer[i],TokenBuffer[i+1]);
		}
		CamConfigSendFIFOCmds();
	}
	else if (ReceivedCmd == EnableTrackingCmd)
	{
		/* publish the event...again with a short delay */
		//Utility_delay(100);
		PlaceEvent(EV_ENABLE_TRACKING);
	}
	else if (ReceivedCmd == DisableTrackingCmd)
	{
		PlaceEvent(EV_DISABLE_TRACKING);
	}
	else if (ReceivedCmd == SetColorMapCmd)
	{
		/* copy the received tokens into the color map */
	  for (i=0; i<TokenCount; i++)
	  {
	    ColorMap[i] = TokenBuffer[i+1];
            /*HexByte(TokenBuffer[i+1], ASCIIBuffer);
            UARTTransmitByte(ASCIIBuffer[0]);
            UARTTransmitByte(ASCIIBuffer[1]);
            UARTTransmitByte(' ');
            if (i==15 || i == 31)
            {
              UARTTransmitByte('\r');
              UARTTransmitByte('\n');
            }*/

            /* write each colorMap byte to EEPROM, but only those
            that changed...this will help reduce wear on the EEPROM */
            eepromData = EEPROMReadByte(i+1);
            if (eepromData != (ColorMap[i]))
            {
                /* need to actually perform the write because the
                data in eeprom is different than the current colorMap */
                EEPROMWriteSucceeded = FALSE;
                while(EEPROMWriteSucceeded == FALSE && NumWrites < MAX_EEPROM_WRITE_ATTEMPTS)
                {
                    EEPROMWriteByte((i+1),ColorMap[i]);
                    NumWrites++;
                    eepromData = EEPROMReadByte(i+1);
                    if (eepromData == ColorMap[i])
                    {
                        EEPROMWriteSucceeded = TRUE;
                    }
                }
                NumWrites = 0;
            }
	  }

#if	DEBUG_COLOR_MAP
    /* for debugging...send out the entire color map */
    PlaceDataInTxBuffer("\r\n",2);
    for (i=0; i<NUM_ELEMENTS_IN_COLOR_MAP; i++)
    {
      MemSet(ASCIIBuffer,0x00,5);
      HexByte(ColorMap[i],ASCIIBuffer);
      PlaceDataInTxBuffer(ASCIIBuffer,2);
      PlaceDataInTxBuffer(" ",1);
      if (i==15 || i == 31)
      {
        /* break up the output */
        PlaceDataInTxBuffer("\r\n",2);
      }
    }
#endif
  }
}

/***********************************************************
	Function Name: UIMgr_processReceivedData
	Function Description: This function is responsible for
	parsing any serial data waiting in the rx fifo
	Inputs:  none
	Outputs: none
***********************************************************/
void CommandProcessReceivedData(void)
{
  unsigned char TempData = 0;

  /* still need to add a mechanism to handle token counts
  that are excessive!!! FIX ME!!! */

  while(IS_DATA_IN_RX_FIFO() == TRUE)
  {
    TempData = CommandReadRxFifo();
    if (TempData == '\r')
    {
      /* we have reached a token separator */
      if (TokenCount == 0)
      {
        /* convert the command */
        ConvertTokenToCommand();
      }
      else
      {
        /* convert a value */
        ConvertTokenToValue();
        TokenCount++;
      }
      /* either way, it is time to try to process the received
      token list since we have reached the end of the cmd. */
      //Utility_delay(100);
      if ((ReceivedCmd == InvalidCmd) || (ReceivedCmd == NoCmd))
      {
        CommandSendNck();
        PlaceEvent(EV_SERIAL_DATA_PENDING_TX);
      }
      else
      {
        CommandSendAck();
        /* publish the serial data pending event, so it
        will push the ACK out before we execute the cmd */
        PlaceEvent(EV_SERIAL_DATA_PENDING_TX);
        ExecuteCommand();
      }
      /* reset any necessary data */
      TokenCount = 0;
      MemSet(TokenBuffer,0x00,MAX_TOKEN_COUNT);
    }
    else if (TempData == ' ')  /* space char */
    {
      /* the end of a token has been reached */
      if (TokenCount == 0)
      {
        ConvertTokenToCommand();
        TokenCount++;   /* check this...why is this being incremented here??? This
                        means we have received a token, with tokenCount == 0, which means it is a
                        command...why is this contributing to tokenCount?
                        This might cause the set color map command to include too much data, since
                        it sets the color map based on tokenCount...CHECK*/
      }
      else
      {
        /* check to see if this token is going to push
        us over the limit...if so, abort the transaction */
        if (TokenCount+1 >= MAX_TOKEN_COUNT)
        {
          /* we received too many tokens, and
          need to NCK this request, since its too
          large...reset everything...*/
          CharCount=0;
          CharIndex=0;
          TokenCount=0;
          ReceivedCmd = InvalidCmd;
        }
        else
        {
          /* tokenCount is still in range...*/
          ConvertTokenToValue();
          TokenCount++;
        }
      }
    }
    else if (((TempData >= 'A') && (TempData <= 'Z')) || ((TempData >= '0') && (TempData <= '9')))
    {
      /* a valid range of token was received */
      ASCIITokenBuffer[CharIndex] = TempData;
      CharCount++;
      CharIndex++;
      if (CharCount > MAX_TOKEN_LENGTH)
      {
        /* we have received a token that cannot be handled...
        set the received cmd to an invalid cmd, and wait
        for the \r to process it */
        ReceivedCmd = InvalidCmd;
        CharIndex = 0;  /* ...so we won't overwrite memory */
      }
    }
    else
    {
      /* an invalid character was received */
      ReceivedCmd = InvalidCmd;
    }
  }  /* end while */
  asm("clt");  /* clear out the T flag in case it wasn't
               cleared already */
}

/***********************************************************
	Function Name: UIMgr_dispatchEvent
	Function Description: This function is responsible for
	processing events that pertain to the UIMgr.
	Inputs:  event - the generated event
	Outputs: none
***********************************************************/
void CommandDispatchEvent(unsigned char event)
{
	switch(event)
	{
		case EV_ACQUIRE_LINE_COMPLETE:
			CommandTransmitPendingData();
			break;

		case EV_SERIAL_DATA_RECEIVED:
			CommandProcessReceivedData();
			break;

		case EV_SERIAL_DATA_PENDING_TX:
			CommandFlushTxBuffer();
			break;
	}
}

/***********************************************************
	Function Name: UIMgr_writeBufferToTxFifo
	Function Description: This function is responsible for
	placing "length" bytes into the tx FIFO.
	Inputs:  pData -  a pointer to the data to send
	         length - the number of bytes to send
	Outputs: none
***********************************************************/
void WriteBufferToTxFifo(unsigned char *pData, unsigned char Length)
{
	unsigned char TempHead;
	if (Length == 0)
	{
		return;
	}

	DisableInterrupts();
	while(Length-- != 0)
	{
		CommandTxFifo[CommandTxFifoHead] = *pData++;

		/* now move the head up */
		TempHead = (CommandTxFifoHead + 1) & (UI_MGR_TX_FIFO_MASK);
		CommandTxFifoHead = TempHead;
	}
	EnableInterrupts();
}




